from paste.script import templates

class BenriTemplate(templates.Template):
    summary = 'Default benri service on the web.'
    _template_dir = 'benri'
    egg_plugins = ['benri']

# TODO: fix for upcoming release
#class ErlangTemplate(templates.Template):
#    summary = 'Creates an Erlang project template.'
#    _template_dir = 'erlang'
#    egg_plugins = ['erlang']



# -*- coding: iso-8859-15 -*-
# (c) Mikael Högqvist

from paste.httpexceptions import HTTPBadRequest

class WSGIApp(object):

    def routes_get(self):
        """Returns a list compatible with `selector.slurp`.
        """
        if hasattr(self, '_routes'):
            return [(key, self._routes[key]) for key in self._routes.keys()]

        # probably a filter app if routes have not been defined
        return []

    def routes_set(self, value):
        self._routes = value

    def routes_del(self):
        del self._routes
                
    routes = property(routes_get, routes_set, routes_del, routes_get.__doc__)

    def _check_arg(self, key, args):
        try:
            return args[key]
        except KeyError, e:
            raise HTTPBadRequest('The argument %s was not specified.' % key)

    def _check_content_size(self, env):
        try:
            return int(env['CONTENT_LENGTH'])
        except KeyError, ValueError:
            raise HTTPBadRequest('Content-Size header not defined or not an integer.')

    def _check_content_type(self, env):
        try:
            return env['CONTENT_TYPE']
        except KeyError, e:
            raise HTTPBadRequest('The Content-Type header was not defined.')
    



# -*- coding: iso-8859-15 -*-
# (c) Mikael Högqvist, ZIB, AstroGrid-D
# This software is licensed under the software license specified at
# http://www.gac-grid.org/

import time, urllib2, base64, os, sys
from threading import Thread

class AuthorizationService(object):
    """An authorization service runs in a separate thread to keep the list
       of users and their optional resource mappings up to date.
    """
    def __init__(self, interval=3600.0):
        """Initializes a SecurityService.
        
        - `interval`: indicates how often the authorization data is updated.
        """

        self.auths = []
           
        self.interval = interval
        self.__running = False
        self.t = Thread(target=self._update)
        self.t.setDaemon(True)

    def _update(self):
        while self.__running:
            for a in self.auths:
                a.update()
            time.sleep(self.interval)

    def addAuth(self, auth):
        """Add an `Authorization`-class to the service.
        
        - `auth`: An `Authorization`-class instance
        """
        if isinstance(auth, Authorization):
            self.auths.append(auth)
        
    def stop(self):
        self.__running = False

    def start(self):
        self.__running = True
        self.t.start()
                                
    def isAuthorized(self, dn, resource=None):
        """Checks if a given user is authorized.
           
        @param dn string with X.509 subject DN
        """
        if len([True for auth in self.auths if auth.isAuthorized(dn, resource=resource)]) > 0:
            return True
        
        return False

# improvements:
# 1. it should be up to the class inheriting from Authorization how often the
# state should be updated.
# 2. a list is inefficient here, problem is how to catch the CN added by proxy 
#    certs when using a hash-map.


# make a dn object and override __eq__?

class Authorize(object):
    def __init__(self, validusers=[]):
        # list of X.509 DN subject strings representing users
        self.validusers = set(validusers)

    def isAuthorized(self, dn, resource=None):
        """Checks if a given user is authorized to access the service
           
        @param dn string with X.509 subject DN
        """
        for user in self.validusers:
            if dn.startswith(user):
                return True

        return False

    def replace(self, users):
        """Replace the current list of valid users
        
        @param users list of X.509 DN strings representing valid users
        """
        self.validusers = set(users)

    def merge(self, users):
        self.validusers = self.validusers.union(users)
                
# str.startsWith()

# this is a class that can be used if the authorization must be done
# at a lower layer in the application, for example if the user auth. must
# be done atomically together with a db write.

class AlwaysAuthorized(object):
    def isAuthorized(self, dn, resource=None):
        return True
        
class AuthorizeVOMRS(Authorize):

    def __init__(self, service, user=None, password=None):
        self.service = service
        self.authstr = None

        if user != None and password != None and user != '' and password != '':
            self.authstr = base64.encodestring('%s:%s' % (user, password))[:-1]

        # timestamp for when last update occurred
        self.ts = 0.0
        
        # how often the service should be asked for a new list of users
        Authorize.__init__(self)

    def _fetch(self):
        req = urllib2.Request(self.service)
        authheader =  "Basic %s" % self.authstr
        req.add_header("Authorization", authheader)
        
        try:
            users = urllib2.urlopen(req)
            self.parse(users)
            self.ts = time.time()
        except IOError, e:
            log.debug("Failed to fetch user list from VOMRS service %s: %s", self.service, str(e))
        except:
            raise

    def parse(self, users):
        self.merge([(user.strip()).strip('"') for user in users])

    def update(self):
        self._fetch()
                
class AuthorizeFlatFile(Authorize):
    """Handles files with one X.509 DN/row.
    
    Format: <DN>\n
            A line starting with a #-character is ignored.
    """
    def __init__(self, path):
        self.path = path
        self.__last_ts = sys.maxint
        Authorize.__init__(self)

    def parse(self, path):
        f = file(path, "r")
        
        self.merge([user.strip() for user in f])
        f.close()
    
    def update(self):
        mtime = os.path.getmtime(self.path)
        if self.__last_ts < mtime:
            self.parse(self.path)
            self.__last_ts = mtime



# -*- coding: iso-8859-15 -*-
# (c) Mikael Högqvist

import selector, signal, os, logging, sys

from threading import Thread
from paste.httpexceptions import HTTPExceptionHandler

#from benri.wsgi import ContentTypeDispatcher

from benri.wsgiserver import CherryPyWSGIServer
from benri.wsgiserver.ssl_pyopenssl import pyOpenSSLAdapter
from benri.log import initLogger

class InitFailure(Exception): pass

class Application(object):
    
    def __init__(self):
        self._wsgi_app = None
        self._apps = []
        self._app = None
            
    def add(self, app):
        self._apps.append(app)
    
    def replace(self, app):
        # this is used when applying filters by replacing the current app
        self._app = app
        
    def fixate(self):
        # use fixate when the middleware is composed
        # this will load all apps into the dispatcher
    
        if len(self._apps) > 0:
            dispatcher = selector.Selector()

            # add the any_noun pattern which consumes anything up to the ;
            dispatcher.parser.patterns['any_noun'] = '[^;]+'
            
            for app in self._apps:
                dispatcher.slurp(app.routes)
        
            self._app = dispatcher

    @property
    def application(self):
        return self._app
            
class Service(object):

    def __init__(self, config, server_threads=20):
        # takes care of benri/server-specific initialization
        # defined in the [benri] and [server] section of the cfg-file

        try:
            self.__server_threads = int(config["server"]["threads"])
        except (KeyError, ValueError):
            self.__server_threads = server_threads
        
        bind = config["server"]["bind"].split(':')        
        self.bind = bind[0], int(bind[1])

        self.server = None
        
        initLogger('benri')
        self.__log = logging.getLogger('benri')

        self.wsgi_app = None
        
        # install a signal handler for a clean shutdown when a kill is
        # received 
        signal.signal(signal.SIGTERM, self._sigterm)

    def useApplication(self, app):
        self.wsgi_app = app
        
    def _sigterm(self, signum, frame):
        raise SystemExit("Received signal SIGKILL")
                            
    def start(self):
        wsgi_app = self.wsgi_app #HTTPExceptionHandler(self.wsgi_app)
                    
        self.server = CherryPyWSGIServer(self.bind, wsgi_app, numthreads=self.__server_threads)
        self.__log.info("HTTP server started at " + self.bind[0] + ':' + str(self.bind[1]))
         
        self.server.start()

    def stop(self):
        if self.server:
            self.server.stop()
            self.__log.info("HTTP server shutdown complete")

class SecureService(object):

    def __init__(self, config, server_threads=20):
        # takes care of benri/server-specific initialization
        # defined in the [benri] and [server] section of the cfg-file

        self.__server_threads = server_threads
        
        bind = config["server"]["bind"].split(':')
        self.bind = bind[0], int(bind[1])

        self.server = None
        
        initLogger('benri')
        self.__log = logging.getLogger('benri')
        
        self.wsgi_app = None

        try:
            self.key_path = config["server"]["ssl_key_path"]
            self.cert_path = config["server"]["ssl_cert_path"]
        except KeyError, ke:
            raise InitFailure('Could not configure the secure server. Missing %s in the [server]-section.' % (str(ke)))

        try:
            self.ca_certs = config["server"]["ssl_cacert_path"]
        except KeyError, ke:
            self.ca_certs = None
            self.__log.info('Client certificates are not verified. Specify ssl_cacert_path in the [server]-section to enable client certificate verification.')

        try:
            self.verify_depth = int(config["server"]["ssl_verify_depth"])
        except KeyError, ke:
            self.verify_depth = 9

        # install a signal handler for a clean shutdown when a kill is
        # received 
        signal.signal(signal.SIGTERM, self._sigterm)

    def useApplication(self, app):
        self.wsgi_app = app
        
    def _sigterm(self, signum, frame):
        raise SystemExit("Received signal SIGKILL")
                            
    def start(self):
        wsgi_app = self.wsgi_app #HTTPExceptionHandler(self.wsgi_app)
        
        self.https_server = CherryPyWSGIServer(self.bind, wsgi_app, numthreads=self.__server_threads)

#        self.https_server.ssl_private_key = self.key_path
#        self.https_server.ssl_certificate = self.cert_path
#        self.https_server.ssl_cacert_path = self.ca_certs
#        self.https_server.ssl_verify_depth = self.verify_depth

        self.https_server.ssl_adapter = pyOpenSSLAdapter(
            self.cert_path, 
            self.key_path, 
            getattr(self, 'ssl_certificate_chain', None),
            client_CA=self.ca_certs,
            client_check_host=True,
            client_check="required")

        self.__log.info("HTTPS server started at " + self.bind[0] + ':' + str(self.bind[1]))
            
        self.https_server.start()

    def stop(self):
        if self.https_server:
            self.https_server.stop()
            self.__log.info("HTTPS server shutdown complete")



"""A library for integrating Python's builtin ``ssl`` library with CherryPy.

The ssl module must be importable for SSL functionality.

To use this module, set ``CherryPyWSGIServer.ssl_adapter`` to an instance of
``BuiltinSSLAdapter``.
"""
import socket, errno

try:
    import ssl
except ImportError:
    ssl = None

from benri import wsgiserver

def decode_cert( prefix, cert ):

    if not cert:
        return None

    key_map = { 'countryName':'C',
                'stateOrProvinceName':'ST',
                'localityName':'L',
                'organizationName':'O',
                'organizationalUnitName':'OU',
                'commonName':'CN',
                # Don't know by what key names python's ssl
                # implementation uses for these fields.
                #'???':'T',
                #'???':'I',
                #'???':'G',
                #'???':'S',
                #'???':'D',
                #'???':'UID',
                'emailAddress':'Email',
                }

    DN_string = ["subject=",]
    cert_dict = {}

    for rdn in cert:
        for key, item in rdn:
            if key in key_map:
                cert_dict["%s_%s" % ( prefix, key_map[key] ) ] = item
                DN_string.append( "%s=%s" % ( key_map[key], item ))

    cert_dict[prefix] = "/".join( DN_string )
        
    return cert_dict

class BuiltinSSLAdapter(wsgiserver.SSLAdapter):
    """A wrapper for integrating Python's builtin ssl module with CherryPy."""
    
    certificate = None
    """The filename of the server SSL certificate."""
    
    private_key = None
    """The filename of the server's private key file."""
    
    def __init__(self, certificate, private_key, certificate_chain=None, client_CA=None):
        if ssl is None:
            raise ImportError("You must install the ssl module to use HTTPS.")
        self.certificate = certificate
        self.private_key = private_key
        self.certificate_chain = certificate_chain
        self.client_CA = client_CA or config.get("server.ssl_client_CA")
        
        self.check_host = config.get("server.ssl_client_check_host", False)
        check = config.get("server.ssl_client_check", "ignore")
        if check == "ignore":
            self.check = ssl.CERT_NONE
        elif check == "optional":
            self.check = ssl.CERT_OPTIONAL
        elif check == "required":
            self.check = ssl.CERT_REQUIRED
        else:
            raise ValueError("server.ssl_client_check must be one of 'ignore','optional','required'")
    
    def bind(self, sock):
        """Wrap and return the given socket."""
        return sock
    
    def wrap(self, sock):
        """Wrap and return the given socket, plus WSGI environ entries."""
        try:
            if self.client_CA:
                s = ssl.wrap_socket(sock, do_handshake_on_connect=True,
                                    server_side=True, certfile=self.certificate,
                                    keyfile=self.private_key, ssl_version=ssl.PROTOCOL_SSLv23,
                                    ca_certs=self.client_CA,
                                    cert_reqs=self.check)
            else:
                s = ssl.wrap_socket(sock, do_handshake_on_connect=True,
                                    server_side=True, certfile=self.certificate,
                                    keyfile=self.private_key, ssl_version=ssl.PROTOCOL_SSLv23)
        except ssl.SSLError, e:
            if e.errno == ssl.SSL_ERROR_EOF:
                # This is almost certainly due to the cherrypy engine
                # 'pinging' the socket to assert it's connectable;
                # the 'ping' isn't SSL.
                return None, {}
            elif e.errno == ssl.SSL_ERROR_SSL:
                if e.args[1].endswith('http request'):
                    # The client is speaking HTTP to an HTTPS server.
                    raise wsgiserver.NoSSLError
            raise
        return s, self.get_environ(s)
    
    # TODO: fill this out more with mod ssl env
    def get_environ(self, sock):
        """Create WSGI environ entries to be merged into each request."""
        cipher = sock.cipher()
        ssl_environ = {
            "wsgi.url_scheme": "https",
            "HTTPS": "on",
            'SSL_PROTOCOL': cipher[1],
            'SSL_CIPHER': cipher[0]
##            SSL_VERSION_INTERFACE 	string 	The mod_ssl program version
##            SSL_VERSION_LIBRARY 	string 	The OpenSSL program version
            }
        
        client_cert = sock.getpeercert()
        if self.client_CA and (client_cert or self.check == ssl.CERT_REQUIRED):
            if self.check == ssl.CERT_REQUIRED and not client_cert:
                sock.close()
                raise ssl.SSLError(ssl.SSL_ERROR_SSL, "certificate required but not found")
            
            client_cert_subject = decode_cert( "SSL_CLIENT_S_DN", client_cert["subject"] )
            if self.check_host:
                try:
                    assert self.address_matches(sock.getpeername(),
                                    client_cert_subject["SSL_CLIENT_S_DN_CN"])
                except:
                    sock.close()
                    raise ssl.SSLError(ssl.SSL_ERROR_SSL, "certificate commonName doesn't match client socket address")
            
            # Update for client environment variables
            ssl_environ.update( {
                #"SSL_CLIENT_M_VERSION":              "",
                #"SSL_CLIENT_M_SERIAL":               "",
                #"SSL_CLIENT_V_START":	         "",
                "SSL_CLIENT_V_END":	                 client_cert["notAfter"],
                #"SSL_CLIENT_A_SIG":	                 "",
                #"SSL_CLIENT_A_KEY":	                 "",
                #"SSL_CLIENT_CERT":	                 "",
                #"SSL_CLIENT_CERT_CHAIN":	         "",
                #"SSL_CLIENT_VERIFY":	         "",
                } )

            ssl_environ.update( client_cert_subject )

            # Update for server environment variables
            ssl_environ.update( {
                #"SSL_SERVER_M_VERSION":              "",
                #"SSL_SERVER_M_SERIAL":               "",
                #"SSL_SERVER_V_START":	         "",
                #"SSL_SERVER_V_END":	                 "",
                #"SSL_SERVER_A_SIG":	                 "",
                #"SSL_SERVER_A_KEY":	                 "",
                #"SSL_SERVER_CERT":	                 "",
                } )

        return ssl_environ
    
    def makefile(self, sock, mode='r', bufsize=-1):
        return wsgiserver.CP_fileobject(sock, mode, bufsize)




# -*- coding: iso-8859-15 -*-
# (c) Mikael Högqvist

import os
from ConfigParser import SafeConfigParser

class ConfigFileMissing(Exception): pass
class RequiredAttrMissing(Exception): pass
class PathNotFound(Exception): pass

class Config(object):
    """
    Config-class that translates a config-file into a dictionary. A config-file
    conforms to the Python config-syntax:
    
    [section]
    attribute = value
    
    The config-dictionary is accessed with:
    config[section][attribute]
    
    Attributes not part of a section are ignored. There are also attributes 
    treated differently depending on their ending:
     
    '_path' - should be a path in the file-system, translated to an absolute path 
              and checked for existance. Raises PathNotFound-exception when the path
              does not exist.
    '_service' - Service-attributes represents a class handling WSGI-requests. The
                 config-dictionary will contain a list with all services.
                 
    - `configpath` - path to the configuration file
    - `basepath` - absolute path used as prefix for relative paths in the config
    - `required_attrs` - a list of (section, attribute)-tuples indicating that
                         the (section, attribute)-pair must exist
    """
    def __init__(self, configpath, workdir, required_attrs=[]):
        config = SafeConfigParser()
        
        if config.read([configpath]) == []:
            raise ConfigFileMissing("Configuration file: " + configpath + " not found")

        self.__data = {}
        
        for sec in config.sections():
            self.__data[sec] = {}
            
            for (item,value) in config.items(sec):
                if item.endswith('_path'):
                    if value.startswith('./'):
                        value = value[2:]
                    value = os.path.join(workdir, value)
                    
                    #if not os.path.exists(value):
                    #    raise PathNotFound(value)
                        
                #elif item.endswith('_service'):
                    # check if the value is a valid class-type
                    
                self.__data[sec][item] = value

        for (sec, attr) in required_attrs:
            try:
                self.__data[sec][attr]
            except KeyError, ke:
                raise RequiredAttrMissing('The required (section, attr)-pair could not be found: (%s, %s)' %(sec, attr))

    def __getitem__(self, key):
        return self.__data[key]

    def keys(self):
        return self.__data.keys()
        
# creates a Singleton-like object out of config
# better than passing around an instance to all
# objects that needs the values in the config-file.
#def Config(configpath, workdir):
#    return _Config(configpath, workdir)
    



# -*- coding: iso-8859-15 -*-
# (c) Mikael Högqvist, ZIB

import random, os

from bsddb import db

READ = 1
WRITE = 2
# owner is allowed to both read and write
OWNER = 3

class UserDB(object):
    """
    The user database persistently stores ACLs for keys. An ACL is a list of
    users and rights mapped to a key indicating who and in what ways a user
    can access the key. External usernames are mapped to a local ID to simplify
    management when a user wants to change their external name.
    
    The rights available in the system are:
    - owner (o): allowed both read and write-access. An owner can also transfer
             ownership to another user. Multiple owners are allowed.
    - read  (r): user is allowed to read data stored at the key
    - write (w): user is allowed to write/change the data store with the key              
    """
    
    # can we assign roles?
    
    def __init__(self, db_path, db_env=None):
        """
        Initializes a user DB.
        
        - `db_path`: path to where the persistent user-data is stored.
        - `db_env`: A Berkeley DB database environment instance (`bdb.db.DBEnv`)
                    Allows external modules using BDB to perform atomic 
                    operations together with the user db.
        """
        
        self.__db_path = db_path
        
        # <External ID> -> <Local ID>
        self.__extid_index = None
        # <Local ID> -> <External ID>
        self.__lid_index = None
        # <<Local ID> + <Key>> -> <rigths>
        self.__acl_lid_index = None
        # <<Key> + <Local ID>> -> <rigths>
        self.__acl_key_index = None

        # when db_env is defined, there is already a directory where the data
        # is stored                
        if db_env == None and not os.path.exists(db_path):
            os.mkdir(db_path)

        self.__init_db(db_path, db_env)

        self.__sep = '|'
        
    def __init_db(self, db_path, db_env):
        # set the initial cache size, 64MB
        self.__db_env = db_env
        self.__ext_db_env = False
        
        if not self.__db_env:
            self.__db_env = db.DBEnv()

            self.__db_env.set_cachesize(0, 2**26)
        
            flags = db.DB_CREATE | db.DB_INIT_MPOOL | db.DB_INIT_LOCK | db.DB_INIT_LOG | db.DB_INIT_TXN
            self.__db_env.open(db_path, flags)
        else:
            self.__ext_db_env = True
            
        # <External ID> -> <Local ID>
        self.__extid_index = db.DB(self.__db_env)        
        self.__extid_index.open("extid_lid.db", None, db.DB_HASH, db.DB_CREATE | db.DB_AUTO_COMMIT)

        # <Local ID> -> <External ID>
        self.__lid_index = db.DB(self.__db_env)
        self.__lid_index.open("lid_extid.db", None, db.DB_RECNO, db.DB_CREATE | db.DB_AUTO_COMMIT)

        # <<Local ID> + <Key>> -> <rigths>
        self.__acl_lid_index = db.DB(self.__db_env)
        self.__acl_lid_index.open("acl_lid.db", None, db.DB_BTREE, db.DB_CREATE | db.DB_AUTO_COMMIT)

        # <<Key> + <Local ID>> -> <rigths>
        self.__acl_key_index = db.DB(self.__db_env)
        self.__acl_key_index.open("acl_key.db", None, db.DB_BTREE, db.DB_CREATE | db.DB_AUTO_COMMIT)

    def create_lid(self, user, txn):
        """Increments the recno, and inserts the new value into the user to 
           localid index. Returns the new localid as a string."""
        lid = "%s" % self.__lid_index.append(user, txn)
        self.__extid_index.put(user, lid)
        return lid
        
    def assign_rights(self, key, user, access_type=READ, txn=None):
        """
        Assigns access rights to a key. As owner, the user is allowed to
        both read and write to the key as well as change owner.
        
        - `key`: the key to which an owner will be assigned.
        - `user`: the user-name of the user
        - `access_type`: defines the access rights allowed to the user
        - `txn`: (Optional), if a txn is given, the operations accessing the
                 db-indexes will be part of the transaction. Otherwise, a new
                 transaction will be created.
        """
        ext_txn = True
        if not txn:
            ext_txn = False
            txn = self.__db_env.txn_begin()

        try:            
            lid = self.__extid_index.get(user)
            
            if not lid:
                # user was not found, create a new local id for the user        
                lid = self.create_lid(user, txn)

            self.__acl_lid_index.put(self.__sep.join([lid, key]), str(access_type), txn=txn)
            self.__acl_key_index.put(self.__sep.join([key, lid]), str(access_type), txn=txn)
            
            if not ext_txn:
                txn.commit(0)
        except:
            if ext_txn:
                raise
            
            txn.abort()
            # TODO: improve error handling
            raise
    
    def remove(self, key, txn=None):
        """
        Remove a key from the index.
        
        - `key`: key to remove
        - `txn`: (Optional), if a txn is given, the operations accessing the
                 db-indexes will be part of the transaction. Otherwise, a new
                 transaction will be created.
        """
        ext_txn = True
        if not txn:
            ext_txn = False
            txn = self.__db_env.txn_begin()

        cursor = self.__acl_key_index.cursor(txn=txn)
        current = cursor.set_range(key)

        try:
            # use a cursor to remove all keys, while registering reverse keys
            while current:
                try:
                    key_lid, value = current
                    
                    curr_key, curr_lid = tuple(key_lid.split(self.__sep))

                    if curr_key != key:
                        current = None
                    else:
                        cursor.delete()
                        self.__acl_lid_index.delete(self.__sep.join([curr_lid, curr_key]), txn=txn)
                        current = cursor.next()
                except:
                    cursor.close()
                    if not ext_txn:
                        txn.abort()
                    raise

            cursor.close()
            
            #self.__acl_lid_index.remove(self.__sep.join([lid, key]), str(access_type), txn=txn)
            #self.__acl_key_index.put(self.__sep.join([key, lid]), str(access_type), txn=txn)

            if not ext_txn:
                txn.commit(0)
        except:
            if ext_txn:
                raise
            
            txn.abort()
            raise
                                    
    def is_authorized(self, key, user, access_type=READ, txn=None):
        """
        Checks if the user is authorized to perform the access method on the 
        key. This can also be used to check if a user is the key-owner by 
        defining `access_type` to `benri.userdb.OWNER`.

        - `key`: key for which the user's rights will be checked
        - `user`: the user-name of the user
        - `access_type`: indicates the type of access, defaults to Read.
        - `txn`: (Optional), if a txn is given, the operations accessing the
                 db-indexes will be part of the transaction. Otherwise, a new
                 transaction will be created.
        """
        ext_txn = True
        if not txn:
            ext_txn = False
            txn = self.__db_env.txn_begin()

        try:
            lid = self.__extid_index.get(user)
                        
            if not lid:
                if not ext_txn:
                    txn.abort()
                return False
            
            rights = int(self.__acl_lid_index.get(self.__sep.join([lid, key])))
                
            if (rights & access_type):
                if not ext_txn:
                    txn.commit(0)
                return True

            if not ext_txn:
                txn.abort()
            return False
        except:
            if ext_txn:
                raise
            
            txn.abort()
            raise

    def user_keys(self, user):
        pass

    def close(self):
        self.__extid_index.close()
        self.__lid_index.close()
        self.__acl_lid_index.close()
        self.__acl_key_index.close()
        
        if not self.__ext_db_env:
            self.__db_env.close()




# -*- coding: iso-8859-15 -*-
# (c) Mikael Högqvist

import logging
"""
This module contains wrappers for the python logging-module.

The logger is accessed via:
log = logging.getLogger(logger_name)

"""

FORMAT = '%(asctime)s (%(levelname)s) %(name)s:%(module)s:%(message)s'

def get_logger(name):
    pass
    
def initLogger(logger_name, level=logging.INFO):
    """
    Initializes a logger for the given ``project``. Adds a StreamHandler
    to the logger.
    """

    log = logging.getLogger(logger_name)
    log.setLevel(level)
    
    # setup a basic stream handler
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(logging.Formatter(FORMAT))
    log.addHandler(stream_handler)

def changeLevel(logger_name, level):
    log = logging.getLogger(logger_name)
    log.setLevel(level)
    
def fileLogger(logger_name, filename):
    """
    Adds a file-logger to the given ``logger_name``.
    """
    file_handler = logging.FileHandler(filename)
    file_handler.setFormatter(FORMAT)

    log = logging.getLogger(logger_name)
    log.addHandler(file_handler)    



# -*- coding: iso-8859-15 -*-
# (c) Mikael Högqvist, ZIB

import time
from Queue import Queue, Empty
from sched import scheduler
from threading import Thread

from benri.db import DB, Index
from benri.threadstage import ThreadStageEvent, ThreadStagePoolEvent, EventHandler

class SchedulerHandler(EventHandler):

    def __init__(self, db_path, scheduler, scheduler_event):
        self.__db = DB(db_path)
        self.__index = Index('lifetime.db', self.__db)
        self.__scheduler = scheduler
        self.__events = {} # key -> event
        self.__scheduler_event = scheduler_event
        self.recover()

    def recover(self):
        # the scheduler thread is not started yet, so there is no risk for 
        # conflicts
        for (entry, ttl) in self.__index:
            event = self.__scheduler.enterabs(float(ttl), 0, self.__scheduler_event, [entry])
            self.__events[entry] = event

    def event_update(self, state):
        entry, trigger_time = state

        try:
            # cancel any previous events registered with this entry
            self.__scheduler.cancel(self.__events[entry])
            del self.__events[entry]
        except KeyError, e:
            # entry does not have an assigned event
            pass
        except RuntimeError, e:
            # this happens when a non-active event was cancelled
            pass
        except Exception, e:
            #print self.__events[entry]
            #print e
            pass
        
#        print "event update: ", state
        # what thread is the scheduler executing in, data_stage thread?
        event = self.__scheduler.enterabs(trigger_time, 0, self.__scheduler_event, [entry])
        self.__events[entry] = event
        self.__index[entry] = str(trigger_time)
        
    def event_delete(self, entry):
        try:
            self.__scheduler.cancel(self.__events[entry])
            del self.__events[entry]            
            del self.__index[entry]
        except ValueError, e:
            # if value error is raised, the event has been triggered, but 
            # worker_done or worker_failed has not yet been called, it is ok
            # to ignore this error since a cleanup of the state will be done
            # when worker_done is called
            pass
        except KeyError, e:
            # key error is ok here, just means that there where no events
            # registered for this key
            pass

    # this event is called when the worker is done with his task
    def event_worker_done(self, entry):
        try:
            del self.__events[entry]
            del self.__index[entry]
        # this means that the event was already removed, because event delete
        # was called after the event was executed by the scheduler but before
        # the worker_done message was sent.
        except KeyError, e:
            pass
            #pass

    def event_worker_failed(self, entry):
        # worker failure removes the event from the event lists, however,
        # TODO: the failure should be passed on as an exception to the user
        # then they can react by for example re-inserting the event.
        try:
            del self.__events[entry]
            del self.__index[entry]
        # this means that the event was already removed, because event delete
        # was called after the event was executed by the scheduler but before
        # the worker_done message was sent.
        except KeyError, e:
            pass
        
#        print "event worker_failed: ", entry, time.time()    

    def close(self):
        self.__index.close()
        self.__db.close()

# TODO: change into persistent scheduler and not lifetime
             
class Lifetime(Thread):
    """
    Maintains a lifetime (Time-To-Live) for keys. Calls a user-defined callback
    with the key as argument when the its lifetime has expired. The time 
    resolution is seconds.
    """
    
    # re-design to use the python sched-module
    # single stage handling both memory and persistent state, since a stage is
    # a single thread, we don't need to lock during change of lifetime for a key
    #
    # memory state: - hash-table with key -> scheduler event, used to cancel 
    #                 active events for a key
    #               - the scheduler itself
    #
    # persistent state: key and timestamp indicating when in seconds after 
    #                   epoch the event should be fired. This is used to rebuild
    #                   the hash-table and the scheduler state.
    #
    # worker pool: The event action is a put to a queue taken care of by a set
    #              of worker threads. This is necessary to not delay the
    #              scheduler when the action is blocking.
    
    def __init__(self, db_path, callback):
        Thread.__init__(self)
        # persistent state of the system
        self.__cb = callback
        
        self.__scheduler = scheduler(time.time, time.sleep)
        self.__sched_handler = SchedulerHandler(db_path, self.__scheduler, self.__scheduler_cb)
        self.__data_stage = ThreadStageEvent(self.__sched_handler)
        self.__worker_pool = ThreadStagePoolEvent(self.__cb, self.__data_stage, 15)

    def run(self):       
        self.__worker_pool.start()
        self.__data_stage.start()
        self.__running = True
        while self.__running:
            time.sleep(0.1)
            self.__scheduler.run()

    def stop(self):
        self.__worker_pool.stop()
        self.__data_stage.stop()
        self.__sched_handler.close()
        self.__running = False

    def change_callback(self, cb):
        self.__cb = cb
        self.__worker_pool.change_callback(cb)
        
    def __scheduler_cb(self, entry):
        # insert into the worker pool, make sure that empty entries
        # triggered by the sched-module are ignored
        if entry != None:
            self.__worker_pool.send(entry)          
                
    def __setitem__(self, entry, ttl):
        """
        Updates the ttl for the given entry. If the entry does not exist, it
        is created.
        """
        # current time + ttl is the time when to fire the event
        trigger_t = time.time() + ttl
        self.__data_stage.send(None, 'update', (entry, trigger_t))

    def __delitem__(self, entry):
        """
        Deletes a previously registered ttl for an entry.
        """
        self.__data_stage.send(None, 'delete', entry)



# -*- coding: iso-8859-15 -*-
# (c) Mikael Högqvist

from benri.app import WSGIApp

class AuthFilter(WSGIApp):
    """
    Middleware that returns 401 Unathorized when an X.509 authenticated 
    user is not allowed to access the wrapped application.
    """

    def __init__(self, app, authservice, host=None, port=443, exclude_http_methods=['GET']):
        """
        Wraps a WSGI-application and filters out any requests which are not
        authorized according to an authentication service. Sets the Location-
        header to a SSL/TLS server if the `HTTP_SSL_CLIENT_S_DN`- or 
        `SSL_CLIENT_S_DN`-variable is not set. It is possible to define HTTP 
        methods which should not be filtered using the `exclude_http_methods`-
        argument.
        
        - `app`: the wrapped WSGI application
        - `authservice`: an instance of an `AuthorizationService`.
        - `host`: specifies a host that runs an SSL/TLS-server. Uses the 
                  `HTTP_HOST` environment-variable if `None`.
        - `port`: port where the SSL/TLS-server is running. Defaults to 443.
        - `exclude_http_methods`: list of HTTP methods for which authorization is
                                  unnecessary. Defaults to ['GET'].
        """
        self.app = app
        
        self.auth = authservice
        self.host = host
        self.port = port
        self.exclude_http_methods = exclude_http_methods
        
    def __call__(self, env, response):
        if env['REQUEST_METHOD'] in self.exclude_http_methods:
            return self.app(env, response)
        
        if not self.host:
            host = env['HTTP_HOST'].split(':')[0]

        port = self.port
        
        client_dn = 'HTTP_SSL_CLIENT_S_DN'
        
        try:
            env[client_dn]
        except KeyError:
            client_dn = 'SSL_CLIENT_S_DN'
            
        plain_text = ('Content-Type', 'text/plain')
        if not client_dn in env:
            headers = [('Location', 'https://' + host + ':' + str(port))]
            headers.append(plain_text)
            response('401 UNAUTHORIZED', headers)
            return ['%s is only allowed via HTTPS\r\n' %(env['REQUEST_METHOD'])]
        
        if client_dn in env and self.auth.isAuthorized(env[client_dn]):
            env['benri.ssl_user'] = env[client_dn]
            return self.app(env, response)
        
        log.debug('User %s was not allowed to %s to the resource.' % (env[client_dn], env['REQUEST_METHOD']))
        # user is not allowed to access the service using this certificate
        response('401 UNAUTHORIZED', [plain_text])
        return ['User %s is not authorized to access the service.\r\n' %(env[client_dn])]
            



# -*- coding: iso-8859-15 -*-
# (c) Mikael Högqvist, ZIB, AstroGrid-D

import simplejson

from paste.fileapp import _FileIter

# a json method decorator
#
# json content type: http://www.ietf.org/rfc/rfc4627.txt
# parses and consumes input data when the content-type indicates that it is 
# json data ('application/json') and puts it in the environment under the key
# 'benri.json'
#
# the response is automatically dumped as json if the content-type header is
# set to 'application/json'

JSON_CONTENT_TYPE = 'application/json'
JSON_HEADER = ('Content-Type', JSON_CONTENT_TYPE)

def json(f, name=None, **kwds):
    def wrapped(*args, **kwargs):
        #print "json: wrapped", args, kwargs
        _self = args[0]
        environ = args[1]
        response = args[2]

        json_header = [JSON_HEADER]

        # used to wrap the start response
        class _Wrap(object):
            def __init__(self):
                self.status = None
                self.headers = []
            
            def resp(self, status, headers):
                self.status = status
                self.headers = headers
       
        req_data = {}
       
        try:
            if environ['REQUEST_METHOD'] in ['PUT', 'POST']:
                chunks = []
                
                for chunk in _FileIter(environ['wsgi.input'], size=int(environ['CONTENT_LENGTH']), block_size=2**14):
                   chunks.append(chunk)

                req_data = simplejson.loads(''.join(chunks))
        except KeyError, e:
            response('400 BAD REQUEST', json_header)
            return [simplejson.dumps("No data in request.")]
        except Exception, e:
            response('400 BAD REQUEST', json_header)
            return [simplejson.dumps("Received invalid JSON data. %s" % (str(e)))]
                    
        wrap = _Wrap()
        try:
            environ['benri.json'] = req_data
            ret_list = f(_self, environ, wrap.resp)
        except:
            raise

        try:
            ret_data = ret_list[0]
        # dont add the json header if it is already part of the headers,
        # and dont transform the data
        # or if the content-type is set already
            if 'content-type' in [a.lower() for (a,b) in wrap.headers]:
                response(wrap.status, wrap.headers)
            # ret is always a list with one element
                return [ret_data]
            else:
                response(wrap.status, json_header + wrap.headers)
                return [simplejson.dumps(ret_data)]
        
        # return this when the data is empty
        except IndexError, e:
            response(wrap.status, [('Content-type', 'text/plain')])
            return []
                    
    wrapped.__doc__ = f.__doc__
    return wrapped



# -*- coding: iso-8859-15 -*-
# (c) Mikael Högqvist, ZIB

import random

from threading import Thread
from Queue import Queue, Empty

SHUTDOWN = 0
SELF = 1
OTHER = 2

class EventHandler(object):
    pass
    
#    def __getattribute__(self, name):
#        try:
#            func = getattr(self, "event_%s" % name)
#        except AttributeError:
#            raise
#        else:
#            return func
            
class ThreadStage(object):
    """
    A stage is a single thread with a queue. The main idea with the stage is
    to serialize access to state in order to avoid locking problems with
    concurrent access. This stage class is simple in the way that calls to a
    single user-defined function is serialized.
    
    Usage example:
    
    >>> import time
    >>> h = {'a':0}
    >>> def f(state):
    ...     k, htable = state
    ...     htable[k] += 1
    >>> stage = ThreadStage(f)
    >>> stage.start()
    >>> stage.send_async(('a', h))
    >>> h
    {'a': 0}
    >>> time.sleep(0.01)
    >>> h
    {'a': 1}
    >>> stage.stop()
    """
    def __init__(self, f):
        self.q = Queue(0)
        self.__f = f
        self.__thread = Thread(target=self.worker)

    def worker(self):
        active = True
        
        while active:
            try:
                (sender, state) = self.q.get(timeout=1)
            except Empty:
                continue
            
            if sender == SELF and state == SHUTDOWN:
                active = False
            else:
                self.__f(state)

    def send_async(self, state):
        self.q.put((OTHER, state))

    def send(self, state):
        self.q.put((OTHER, state))
    
    def start(self):
        self.__thread.start()
        
    def stop(self):
        self.q.put((SELF, SHUTDOWN))
        self.__thread.join()

class ThreadStageEvent(object):
    """
    The ThreadStageEvent-class takes an instance that inherits from the
    EventHandler-class and implements event handler methods.
    
    Example usage:
    
    >>> import time
    >>> class C(EventHandler):
    ...     def __init__(self):
    ...         self.state = 0
    ...     def event_test(self, state):
    ...         self.state += state
    ...         return self.state
    >>> c = C()
    >>> stage = ThreadStageEvent(c)
    >>> stage.start()
    >>> stage.send(None, 'test', 5)
    >>> c.state
    0
    >>> time.sleep(0.01)
    >>> c.state
    5
    >>> stage.stop()
    """
    
    def __init__(self, instance):
        self.q = Queue(0)
        self._thread = Thread(target=self.worker)
        self._instance = instance
#        self.__thread.start()        

    def worker(self):
        active = True
        
        while active:
            try:
                (sender, event, state) = self.q.get(timeout=1)
            except Empty:
                continue
            except Exception, e:
                raise

            if sender == SELF and event == SHUTDOWN:
                active = False
            else:
                try:
                    f = getattr(self._instance, "event_%s" % event)
                    f(state)
                except AttributeError:
                    # if the event does not exist, message is lost
                    pass

    def send(self, sender, event, state):
        """
        Sends a message event to the local stage.
        
        `sender` - an instance of the stage that sent the message, this is used
                   for sending a message back with the results.
        `event`  - the event passed to the stage
        `state`  - the state passed together with the event
        """
        self.q.put((sender, event, state))

    def start(self):
        self._thread.start()
        
    def stop(self):
        self.q.put((SELF, SHUTDOWN, None))
        self._thread.join()

class ThreadStageOrderedEvent(ThreadStageEvent):
    """
    The ThreadStageEventOrdered-class extends ThreadStageEvents by executing
    the incoming messages according to their version. A message with version
    v must execute before a message with version v+1. If there are any gaps in
    the versioning, the stage will wait until it receives all messages filling
    the gap. Messages with a version number less than the current are dropped.
    
    Example usage:
    
    >>> import time  
    >>> class C(EventHandler):
    ...     def __init__(self):
    ...         self.state = 0
    ...     def event_test(self, state):
    ...         self.state += state
    ...         return self.state
    >>> c = C()
    >>> stage = ThreadStageOrderedEvent(c)
    >>> stage.start()
    >>> stage.send(None, 'test', 1, 0)
    >>> c.state
    0
    >>> time.sleep(0.01)    
    >>> c.state
    1
    >>> stage.send(None, 'test', 1, 1)
    >>> stage.send(None, 'test', 1, 3)
    >>> time.sleep(0.01)
    >>> c.state
    2
    >>> stage.send(None, 'test', 1, 2)
    >>> time.sleep(0.01)
    >>> c.state
    4
    >>> stage.send(None, 'test', 1, 2)
    >>> time.sleep(0.01)
    >>> c.state
    4
    >>> stage.send_blocking(None, 'test', 1, 4)
    5
    >>> stage.stop()
    """

    def __init__(self, instance):
        ThreadStageEvent.__init__(self, instance)
        self.__running = False
    
    def send(self, sender, event, state, version):
        """
        Sends a message event to the local stage.
        
        `sender` - an instance of the stage that sent the message, this is used
                   for sending a message back with the results.
        `event`  - the event passed to the stage
        `state`  - the state passed together with the event
        `version` - version of the message
        """
        self.q.put((sender, event, state, version, None))

    def send_blocking(self, sender, event, state, version):
        """
        Sends a message event to the local stage. Blocks the calling thread
        until the results are returned. If the called method instance raises
        an exception, this is re-raised here.
        
        `sender` - an instance of the stage that sent the message, this is used
                   for sending a message back with the results.
        `event`  - the event passed to the stage
        `state`  - the state passed together with the event
        `version` - version of the message
        """
        block_q = Queue(0)
        
        self.q.put((sender, event, state, version, block_q))

        while True:

            try:
                results = block_q.get(timeout=0.1)
                    
                if isinstance(results, Exception):
                    raise results

                return results
            except Empty, e:
                pass
                
    def worker(self):
        active = True
        current_version = 0
        waiting_msgs = []
        
        while active:
            try:
                (sender, event, state, version, block_q) = self.q.get(timeout=1)
            except Empty:
                continue
#            except Exception, e:
#                raise

            
            if sender == SELF and event == SHUTDOWN:
                active = False
            else:
                if version == current_version:
                    try:
                        f = getattr(self._instance, "event_%s" % event)
                        ret = f(state)
                        
                        if block_q:
                            block_q.put(ret)
                    except Exception, e:
                        if block_q:
                            block_q.put(e)
                    
                    # re insert all the waiting messages
                    for sender, event, state, msg_version in waiting_msgs:
                        if msg_version > current_version:
                            self.q.put((sender, event, state, msg_version, block_q))
                    
                    waiting_msgs = []
                    current_version += 1
                                            
                elif version > current_version:
                    waiting_msgs.append((sender, event, state, version))

    def start(self):
        self.__running = True    
        self._thread.start()

    def stop(self):
        self.__running = False
        self.q.put((SELF, SHUTDOWN, None, None, None))
        self._thread.join()

class ThreadStagePeriodic(ThreadStage):
    def __init__(self, trigger, interval):
        ThreadStage.__init__(self, trigger)
        self.__interval = interval

    def worker(self):
        active = True
        
        while active:
            try:
                (sender, state) = self.__q.get(timeout=self.__interval)
            except Empty:
                self.__f()
                continue
            
            if sender == SELF and state == SHUTDOWN:
                active = False

# in a thread stage pool, we have several threads sharing a common queue
# i.e. we cant ensure serialized access to state
# this class is just a simple worker pool
class ThreadStagePool(ThreadStage):
    """
    A simple thread pool which executes a user-defined function. Unlike the
    ThreadStage-class, this function may be executed concurrently by any of the
    `num_workers` threads in the thread-pool.
    
    Example usage:
    
    >>> def f(state):
    ...     state + 1
    >>> pool = ThreadStagePool(f, 5)
    >>> pool.start()
    >>> pool.send_async(12)
    >>> pool.stop()
    """
    def __init__(self, f, num_workers):
        ThreadStage.__init__(self, f)
        self.__num_workers = num_workers
        self.__threads = []

        for i in range(self.__num_workers):
            self.__threads.append(Thread(target=self.worker))

    def start(self):
        for t in self.__threads:
            t.start()

    def stop(self):
        for i in range(self.__num_workers):
            self.q.put((SELF, SHUTDOWN))

class ThreadStagePoolEvent(ThreadStage):
    """
    A thread pool which executes a user-defined function and replies to another
    stage. The events passed back are `worker_done` and `worker_failed`. Unlike
    the ThreadStageEvent-class, this function may be executed concurrently by any
    of the `num_workers` threads in the thread-pool.
    
    Example usage:
    
    >>> import time
    >>> def f(state):
    ...     return state + 1
    >>> class Handler(EventHandler):
    ...     def event_worker_done(self, res):
    ...         return res
    ...     def event_worker_failed(self, exc):
    ...         raise exc
    >>> stage = ThreadStageEvent(Handler())
    >>> stage.start()
    >>> pool = ThreadStagePoolEvent(f, stage, 5)
    >>> pool.start()
    >>> pool.send(12)
    >>> time.sleep(0.1)
    >>> pool.stop()
    >>> stage.stop()
    """
    def __init__(self, f, stage, num_workers):
        ThreadStage.__init__(self, f)
        self.__f = f
        self.__num_workers = num_workers
        self.__threads = []
        self.__stage = stage
        
        for i in range(self.__num_workers):
            self.__threads.append(Thread(target=self.worker))

    def worker(self):
        active = True
        
        while active:
            try:
                (sender, state) = self.q.get(timeout=1)
            except Empty:
                continue
            
            if sender == SELF and state == SHUTDOWN:
                active = False
            else:
                try:
                    ret = self.__f(state)
                    self.__stage.send(self, 'worker_done', state)
                except Exception, e:
                    self.__stage.send(self, 'worker_failed', e)

    def change_callback(self, cb):
        self.__f = cb
        
    def start(self):
        for t in self.__threads:
            t.start()

    def stop(self):
        for i in range(self.__num_workers):
            self.q.put((SELF, SHUTDOWN))

if __name__ == '__main__':
    import doctest
    doctest.testmod()



# -*- coding: iso-8859-15 -*-
# (c) Mikael Högqvist, ZIB

import os, thread, cPickle as pickle

from threading import Thread
from Queue import Queue, Empty

try:
    from bsddb3 import db as bdb
except ImportError, e:
    from bsddb import db as bdb

from shutil import rmtree, copytree

BTREE = bdb.DB_BTREE
HASH = bdb.DB_HASH

class TransactionExpired(Exception): pass
class FailedIteration(Exception): pass

# A transaction decorator for BDB (from rdflib)
def transaction(f, name=None, **kwds):
    def wrapped(*args, **kwargs):
        index_db = args[0].db
        retries = 10
        delay = 1
        e = None
        
        #t = kwargs['env'].txn_begin()
        while retries > 0:
            (kwargs['txn'], state) = index_db.atomic_begin()

            try:
                result = f(*args, **kwargs)
                index_db.atomic_commit()
                # returns here when the transaction was successful
                return result
            except MemoryError, e:
                # Locks are leaking in this code or in BDB
                # print "out of locks: ", e, sys.exc_info()[0], self.db_env.lock_stat()['nlocks']
                index_db.atomic_rollback()
                retries = 0
            except bdb.DBLockDeadlockError, e:
                #print "Deadlock when adding data: ", e
                index_db.atomic_rollback()
                #sleep(0.1*delay)
                #delay = delay << 1
                retries -= 1
#            except bdb.DBRunRecoveryError, e:
#                print "DBRunRecoveryError", e
#                raise
            # this is raised when the generator is supposed to finish, 
            # and should be re-raised to indicate that the generator is done
            # its also called on garbage collection (ensured by the decorated
            # methods)
            except StopIteration, e:
                raise StopIteration
            
            except FailedIteration, e:
                # this is raised during an iteration with a cursor when a 
                # deadlock occurred. The correct solution is to rollback the
                # transaction. Then StopIteration is raised to notify a calling
                # application.
                index_db.atomic_rollback()
                raise StopIteration                
            # if it is a generic exception, assume it comes from the user and
            # not from BerkeleyDB, abort transaction and forward to the caller.
            except Exception, e:
#                import traceback
#                traceback.print_exc()
                index_db.atomic_rollback()
                raise
                                
        #print "Retries failed!", bdb.db_env.lock_stat()['nlocks']
        raise TransactionExpired("Add failed after exception:" % str(e))

#        except Exception, e:
#            print "Got exception: ", e            
#            bdb.rollback()
            
            #t.abort()

    wrapped.__doc__ = f.__doc__
    return wrapped

class Index(object):
    def __init__(self, name, db, index_type=BTREE, sep='^'):
        self.__name = name
        self.db = db
        self.__db_env = db.db_env
        self.__index_type = index_type
        
        self.__index = bdb.DB(self.__db_env)
        
        # flags necessary to open a free-threaded index with support for
        # transactions.
        # See: http://www.oracle.com/technology/documentation/berkeley-db/db/ref/transapp/data_open.html
#        if self.db.fast_txn:
#            flags = bdb.DB_CREATE | bdb.DB_THREAD
#        else:
        flags = bdb.DB_CREATE | bdb.DB_AUTO_COMMIT | bdb.DB_THREAD

        self.__index.open(self.__name, None, self.__index_type, flags)
        self.__sep = sep

    def __setitem__(self, key, value):
        @transaction
        def setitem(self, key, value, txn=None):
            self.__index.put(key, value, txn=txn)

        try:
            setitem(self, key, pickle.dumps(value))
        except Exception, e:
            raise e

    def __getitem__(self, key):
        @transaction
        def getitem(self, key, txn=None):
            val = self.__index.get(key, txn=txn)
            if val == None or pickle.loads(val) == None:
                raise KeyError(key)
           
            return pickle.loads(val)
 
        try:
            return getitem(self, key)
        except Exception, e:
            # print "Got exception in _add: ", e
            raise e

    def __delitem__(self, key):
        """Removes the key from the hash table."""
        
        @transaction
        def delitem(self, key, txn=None):
            self.__index.delete(key, txn=txn)
            
        try:
            return delitem(self, key)
        except Exception, e:
            raise e

# TODO: contains

    def __iter__(self):
        @transaction
        def _iter(self, txn):
            cursor = self.__index.cursor(txn=txn)
            current = cursor.first() # set_range(prefix)

            while current:
                try:
                    key, value = current

                    if key:                
                        yield (key, pickle.loads(value))
                        current = cursor.next()
                        #print "next: ", current
                    else:
                        current = None
                # this exception is raised when the user does generator.close()
                except GeneratorExit, e:
                    cursor.close()
                    raise StopIteration
                except bdb.DBLockDeadlockError, e:
                    cursor.close()
                    raise FailedIteration(e)
                except Exception, e:
                    #print e
                    cursor.close()
                
            
            cursor.close()
            raise StopIteration

        try:
            return _iter(self)
        except Exception, e:
            # print "Got exception in _add: ", e
            raise
        
    def range_query(self, prefix):
        """
        Executes a range query only returning the values starting with prefix.
        The caller can stop the generator at any time by calling close() on
        the instance (see PEP-0342 for details).
        """
        @transaction
        def _range_query(self, prefix, txn):
            cursor = self.__index.cursor(txn=txn)
            current = cursor.set_range(prefix)

            while current:
                try:
                    key, value = current

                    if key and key.startswith(prefix):                
                        yield (key, pickle.loads(value))
                        current = cursor.next()
                        #print "next: ", current
                    else:
                        current = None
                # this exception is raised when the user does generator.close()
                except GeneratorExit, e:
                    cursor.close()
                    raise StopIteration
                except bdb.DBLockDeadlockError, e:
                    cursor.close()
                    raise FailedIteration(e)                    
                except Exception, e:
                    #print e
                    cursor.close()

            cursor.close()
            raise StopIteration

        try:
            return _range_query(self, prefix)
        except Exception, e:
            # print "Got exception in _add: ", e
            raise

    def __len__(self):
        return self.__index.stat()['ndata']
        
    def close(self):
        try:
            self.__index.close()
        except bdb.DBRunRecoveryError, e:
            # this can occur when a second process using the index crashes 
            # without closing it correctly, ignore this since recovery
            # is run on start-up           
            pass

class DB(object):
    """
    A wrapper class for Berkeley DB transaction environments and databases.
    Hides complexity of transaction management and threads.
    """
    
    def __init__(self, db_path, main_thread=False, backup_path=None, checkpoint_interval=600.0, backup_interval=3600.0*24.0, cachesize=2**26, fatal_recovery=False, fast_txn=False):
        """
        Initalize the DB environment.
        
        ``db_path`` - base directory of the environment
        ``main_thread`` - indicates that this thread should perform checkpoints,
                          archival operations and logging. See the Berkeley DB
                          documentation on Transactional Stores for further
                          information.
        ``backup_path`` - Path to where the database should be archived. It is
                          recommended to perform backups to a separate disk or
                          to a network file-system mount. This path is used if
                          fatal recovery needs to be performed. If the value is
                          None, no archives will be created and fatal recovery
                          is not possible.
                            
        ``checkpoint_interval`` - Indicates how often in seconds a checkpoint 
                                  should be performed.
        ``backup_interval`` - Indicates how often in seconds the database
                              backup procedure is executed.
        ``cachesize`` - sets the size in bytes of the in-memory cache.
        ``fatal_recover`` - performs fatal recovery of the database
        ``fast_txn`` - increases performance by doing in memory logging 
                           and auto-removal of disk logs. Disables D in ACID.
        """
        self.__db_path = os.path.abspath(db_path)

        # create the directory if the environment does not exist already        
        if not os.path.exists(self.__db_path):
            os.mkdir(self.__db_path)

        # db open uses relative names for log and data-dir
        rel_data_dir = 'db_data'
        self.__data_dir = os.path.join(self.__db_path, rel_data_dir)
        if not os.path.exists(self.__data_dir):
            os.mkdir(self.__data_dir)

        rel_log_dir = 'db_logs'
        self.__log_dir = os.path.join(self.__db_path, rel_log_dir)
        if not os.path.exists(self.__log_dir):
            os.mkdir(self.__log_dir)
        
        #panic_event = lambda 

        # initialize the environment
        self.db_env = bdb.DBEnv()
        self.db_env.set_cachesize(0, cachesize)
        
        # set the data dir
        # http://www.oracle.com/technology/documentation/berkeley-db/db/api_c/env_set_data_dir.html
        self.db_env.set_data_dir(rel_data_dir)
        
        # set the log dir and logging parameters
        self.db_env.set_lg_dir(rel_log_dir)

        # enable BDB deadlock-detection
        self.db_env.set_lk_detect(bdb.DB_LOCK_MAXLOCKS)
        
        # flags necessary to create a transaction-aware environment.
        # see: http://www.oracle.com/technology/documentation/berkeley-db/db/ref/transapp/env_open.html

        recovery_type = bdb.DB_RECOVER
                
        if fatal_recovery:
            recovery_type = bdb.DB_RECOVER_FATAL
        
        if not main_thread:
            recovery_type = 0x0
                    
        flags = bdb.DB_CREATE | bdb.DB_INIT_MPOOL | bdb.DB_INIT_LOCK | bdb.DB_INIT_TXN | bdb.DB_INIT_LOG | bdb.DB_THREAD | recovery_type

        self.fast_txn = fast_txn
#        if fast_txn:
#            flags |= bdb.DB_LOG_INMEMORY | bdb.DB_LOG_AUTOREMOVE | bdb.DB_TXN_WRITE_NOSYNC
#            flags |= bdb.DB_TXN_WRITE_NOSYNC


        if fast_txn:
            fast_txn_flags = bdb.DB_TXN_WRITE_NOSYNC | bdb.DB_LOG_AUTOREMOVE
            self.db_env.set_flags(fast_txn_flags, True)

#            self.db_env.set_lg_
#            self.db_env.set_lg_max(10)
#            print self.db_env.get_lg_max()
#            print "fewbfew"
#            try:
#                fast_txn_flags = bdb.DB_LOG_INMEMORY
#                self.db_env.set_flags(fast_txn_flags, True)
#                self.db_env.set_lg_bsize(1024)
#            except Exception, e:
#                print e

        try:
            self.db_env.open(db_path, flags)
        except Exception, e:
            print "db env could not be opened:", e
            
        
        self.__txns = {}
        
        self.__main_thread = main_thread
        self.__backups_enabled = False
        
        # setup checkpointing and the hot backup thread
        if self.__main_thread:
            self.__shutdown_q = Queue()
            self.__checkpoint_interval = checkpoint_interval
            self.__checkpoint_t = Thread(target=self._checkpoint)
            self.__checkpoint_t.start()
            
            if backup_path != None:
                self.__backup_path = backup_path
                
                if not os.path.exists(self.__backup_path):
                    os.mkdir(self.__backup_path)
                
                self.__backup_interval = backup_interval
                self.__backup_t = Thread(target=self._hot_backup)
                self.__backup_t.start()
                self.__backups_enabled = True

    def _hot_backup(self):
        def backup():
            try:
                # perform the hot backup procedure as described in:
                # http://www.oracle.com/technology/documentation/berkeley-db/db/ref/transapp/archival.html
                # see the source for the db_hotbackup-utility for more details
                
                # 1. checkpoint and remove any uncessary log-files
                # this is to avoid copying extra data
                self.db_env.txn_checkpoint()
                self.db_env.log_archive(bdb.DB_ARCH_REMOVE)
                
                # 2. check existance and rights of the root target directory
                if not os.path.exists(self.__backup_path):
                    raise Exception('Could not perform backup since the target directory does not exist, %s' % self.__backup_path)
                
                backup_dir = os.path.join(self.__backup_path, 'db_backup')

                # 3. clean the target directory, assume that the existing data 
                #    has been secured through other means.
                if os.path.exists(backup_dir):
                    rmtree(backup_dir)
                    # recreate the backup-dir since remove_dirs removes it...
                    os.mkdir(backup_dir)
                
                # 4. Copy the data dir to the target dir
                backup_data_dir = os.path.join(backup_dir, os.path.basename(self.__data_dir))
                copytree(self.__data_dir, backup_data_dir)
                
                # 5. Copy the log-dir
                backup_log_dir = os.path.join(backup_dir, os.path.basename(self.__log_dir))
                copytree(self.__log_dir, backup_log_dir)
                

                # 6. perform a fatal recovery on the backup to ensure that
                #    it is valid 
                
                test_env = bdb.DBEnv()
                test_env.set_lg_dir(os.path.basename(backup_log_dir))
                test_env.set_data_dir(os.path.basename(backup_data_dir))
                test_env.set_lk_detect(bdb.DB_LOCK_MAXLOCKS)
                
                recovery_type = bdb.DB_RECOVER_FATAL
                flags = bdb.DB_CREATE | bdb.DB_INIT_MPOOL | bdb.DB_INIT_LOCK | bdb.DB_THREAD | bdb.DB_INIT_LOG | bdb.DB_INIT_TXN | recovery_type
                test_env.open(backup_dir, flags)                
                test_env.close()
                
            except Exception, e:
                print 'hot backup failed: ', e
                
        while True:
            try:
                value = self.__shutdown_q.get(timeout=self.__backup_interval)
                return
            except Empty, e:
                backup()

    def _checkpoint(self):
        while True:
            try:
                value = self.__shutdown_q.get(timeout=self.__checkpoint_interval)
                # if shutdown queue gets a value, finish the thread
                return
            except Empty, e:
                # this can only fail due to arguments
                # http://www.oracle.com/technology/documentation/berkeley-db/db/api_c/txn_checkpoint.html
                self.db_env.txn_checkpoint()
                
    def atomic_begin(self):
        """
        All db calls made within an atomic block is executed with the same 
        transaction. Nested calls to atomic_begin returns the active 
        transaction.
        """
        
        active = False
        try:
            self.__txns[thread.get_ident()]
            active = True
        except KeyError:
            self.__txns[thread.get_ident()] = self.db_env.txn_begin()
        
        return (self.__txns[thread.get_ident()], active)

    def atomic_rollback(self):
        try:
            txn = self.__txns[thread.get_ident()]
            txn.abort()
            del self.__txns[thread.get_ident()]
            active = False
        except KeyError:
            active = False
        except:
            #txn.abort()
            raise
            
    def atomic_commit(self):
        # ends a currently active atomic block
        try:
            txn = self.__txns[thread.get_ident()]
            txn.commit(0)
            del self.__txns[thread.get_ident()]
            active = False
        except KeyError:
            active = False
        except:
            txn.abort()

    def close(self):
        if self.__main_thread:
            # shutdown the checkpointing and backups, must put two shutdowns
            # here to ensure that both threads finish
            self.__shutdown_q.put(None)
            self.__shutdown_q.put(None)            
            self.__checkpoint_t.join()

            if self.__backups_enabled:
                self.__backup_t.join()        
        try:
            self.db_env.close()
        except bdb.DBRunRecoveryError, e:
            # this can occur when a second process using the index crashes 
            # without closing it correctly, ignore this since recovery
            # is run on start-up
            pass    
                    



# -*- coding: iso-8859-15 -*-
# (c) Mikael Högqvist

import selector, os

from paste.httpexceptions import HTTPException

def daemonize():
    """
    Spawn a daemon process of the current executable. It is recommended to use this
    when running a server-process.
    """
    # do the UNIX double-fork magic, see Stevens' "Advanced
    # Programming in the UNIX Environment" for details (ISBN 0201563177)
    # adapted from http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/278731
    try:
        pid = os.fork()
    except OSError, e:
        print >>sys.stderr, "Daemon could not be started: %d (%s)" % (e.errno, e.strerror)
        sys.exit(1)

    if pid == 0:
        os.setsid()
            
        try:
            pid = os.fork()
        except OSError, e:
            print >>sys.stderr, "Daemon could not be started: %d (%s)" % (e.errno, e.strerror)
            sys.exit(1)

        if pid == 0:               
            os.chdir("/")   #don't prevent unmounting....
            os.umask(0)
        else:
            os._exit(0) # see comments in the recipe
    else:
        os._exit(0)

    import resource
    maxfd = resource.getrlimit(resource.RLIMIT_NOFILE)[1]
    if (maxfd == resource.RLIM_INFINITY):
        maxfd = 1024
      
    for fd in range(0, maxfd):
        try:
            os.close(fd)
        except OSError:
            pass
            
    os.open(os.devnull, os.O_RDWR)	# standard input (0)

    # Duplicate standard input to standard output and standard error.
    os.dup2(0, 1)			# standard output (1)
    os.dup2(0, 2)			# standard error (2)

def test_wrapper(wsgiapp):
    """Returns a selector-wsgiapp wrapping the given wsgiapp instance. Useful
       for testing when using paste fixtures.
    """

    s = selector.Selector()
    s.slurp(wsgiapp.routes)
    
    return ErrorMiddleware(s)



# -*- coding: iso-8859-15 -*-
# (c) Mikael Högqvist

from benri.app.wsgiapp import WSGIApp
from benri.app.wsgicollection import Collection as WSGICollection
from benri.app.json import json, JSON_CONTENT_TYPE



"""
The MIT License

Copyright (c) 2006 Joe Gregorio  

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

NOTE: This code is taken from Joe Gregorio and adapted for benri usage.

WSGICollection

The idea of RESTful "Collections" has been percolating for years now.
A Collection is nothing more than a list, a container for resources.
While the APP_ defines a Collection in terms of Atom Feed and Entry documents
we don't have to be limited to that. It's time to complete a virtuous circle;
RESTLog_ inspired the Atom Publishing Protocol which 
inspired David Heinemeier Hansson's
`World of Resources`_ (pdf) and now it's time to come full circle and get that 
world of resources in Python.

In particular look at page 18 of that slide deck, where
dispatching to a collection of people, the following URIs
are to be handled:

+-------+-----------------+
| GET   |  /people        |
+-------+-----------------+
| POST  |  /people        |
+-------+-----------------+
| GET   |  /people/1      |
+-------+-----------------+
| PUT   |  /people/1      |
+-------+-----------------+
| DELETE|  /people/1      |
+-------+-----------------+
| GET   |  /people;new    |
+-------+-----------------+
| GET   |  /people/1;edit |
+-------+-----------------+


Now the 'new' and 'edit' URIs can be a bit ambiguous, only in the sense
that you might not guess right away that they are nouns, and remember,
URIs always identify nouns. I prefer to make the noun-ishness of them more apparent.

+-----+----------------------+
| GET |  /people;create_form |
+-----+----------------------+
| GET |  /people/1;edit_form |
+-----+----------------------+


In general, using the notation of wsgidispatcher_, we are looking at URIs 
of the form::

       /...people/[{id:word}][;{noun}] 

And dispatching requests to URIs of that form to functions with nice names::
        

  GET    /people               list()
  POST   /people               create()
  GET    /people/1             retrieve()
  PUT    /people/1             update()
  DELETE /people/1             delete()
  GET    /people;create_form   get_create_form()
  GET    /people/1;edit_form   get_edit_form()
  

WSGICollection relies on WSGI middleware before it in the call
chain to parse the URIs for {id} and {noun}, such 
as wsgidispatcher_. In theory it will 
work with any WSGI middleware that sets values for 'id' and 'noun' in 
environ['wsgiorg.routing_args'].
Here is how you would define a WSGI application that implements a collection::
    

    from wsgicollection import Collection
    
    class RecipeCollection(Collection):
    
        # GET /cookbook/
        def list(environ, start_response):
            pass
        # POST /cookbook/
        def create(environ, start_response):
            pass
    
        # GET /cookbook/1
        def retrieve(environ, start_response):
            pass
    
        # PUT /cookbook/1
        def update(environ, start_response):
            pass
    
        # DELETE /cookbook/1
        def delete(environ, start_response):
            pass
    
        # GET /cookbook/;create_form
        def get_create_form(environ, start_response):
            pass
    
        # POST /cookbook/1;comment_form
        def post_comment_form(environ, start_response):
            pass


And this class can be easily hooked up to wsgidispatcher::

    from wsgidispatcher import Dispacher 
    
    urls = Dispatcher()
    urls.add('/cookbook/[{id:word}][;{noun}]', RecipeCollection())

.. _RESTLog: http://bitworking.org/news/RESTLog_Overview    
.. _APP: http://bitworking.org/projects/atom/ 
.. _`World of Resources`: http://www.loudthinking.com/lt-files/worldofresources.pdf
.. _wsgidispatcher: wsgidispatcher.html
"""

import re, sys
from benri.app import WSGIApp
from logging import info, error

COLL_MAP = {
    'GET': 'list',
    'POST': 'create'
    }

ENTRY_MAP = {
    'GET': 'retrieve',
    'PUT': 'update',
    'DELETE': 'delete'
}

class Collection(WSGIApp):
    """
    """
    def __init__(self, uri_prefix='[/]'):
        self._id = "" 
        self._noun = ""
        self._function_name = ""

        if uri_prefix.endswith('/'):
            uri_prefix = uri_prefix[:-1] + '[/]'
        
        self.routes = {uri_prefix + '[{id:any_noun}[;{noun:any}]]': dict(GET=self, POST=self, PUT=self, DELETE=self)}
#                       uri_prefix + '{id:any}': dict(DELETE=self),
#                       uri_prefix + '{id:any}[/{noun:any}]': dict(GET=self, PUT=self)}

    def __call__(self, environ, start_response):
        self._id = "" 
        self._noun = ""
        self._function_name = ""
        if 'wsgiorg.routing_args' in environ:
            url_vars = environ['wsgiorg.routing_args'][1]
        elif 'selector.vars' in environ:
            url_vars = environ['selector.vars']
        else:
            start_response("500 Internal Server Error", [('content-type', 'text/plain')])
            #error("Environment variables for wsgicollection.Collection not provided via WSGI. %s" % str(environ))
            return ['Environment variables for wsgicollection.Collection not provided via WSGI.']

        self._id = url_vars.get('id', '')
        #self._noun = url_vars.get('noun', '')
        method = environ['REQUEST_METHOD']
# Note that we let the user take care of nouns, sometimes it is better to delay the getattr
#        self._function_name = "%s_%s" % (method.lower(), self._noun)
#        if not self._noun:
        method_map = self._id and ENTRY_MAP or COLL_MAP
        self._function_name = method_map.get(method, '') 

        if self._function_name and not self._function_name.startswith("_") and self._function_name in dir(self):
            return getattr(self, self._function_name)(environ, start_response)
        else:
            start_response("404 Not Found", [("Content-Type", "text/plain")])
            return ["Resource not found."]






try:
    from setuptools import setup, find_packages
except ImportError, e:
    from ez_setup import use_setuptools
    use_setuptools()
finally:
    from setuptools import setup, find_packages

setup(name='benri',
    version='0.0.4', 
    author='Mikael Hoegqvist',
    author_email='hoegqvist@zib.de',
    url='',
    download_url='',
    description='A REST-service creation framework.',
    license='',
    long_description="""Convenience framework for REST service creation. Uses
                        Paste for many basic tasks.""",
    include_package_data = True,
    packages = find_packages(),
    install_requires = [
    'benri.client >= 0.0.2',
    'Paste>=1.7.5.1',
    'PasteScript>=1.7.2',
    'selector>=0.8.11',
    ],    
#    benri_erlang=benri.templates:ErlangTemplate
    entry_points = """
    [paste.paster_create_template]
    benri=benri.templates:BenriTemplate
    """,
     classifiers=[
    'Development Status :: 4 - Beta',
    'Environment :: Web Environment',
    'Intended Audience :: Developers',
    'Programming Language :: Python',
    'Topic :: Internet :: WWW/HTTP',
    'Topic :: Software Development :: Libraries',
    ],
    zip_safe = False,
    )




# -*- coding: iso-8859-15 -*-
# (c) Mikael Högqvist

import os

from paste.fileapp import _FileIter
from benri.service import InitFailure
from tempfile import mkstemp

class Spool(object):
    """Spool class which retrieves the input-data via wsgi.input and dumps it in
       a temporary file in a spool-directory. Can be used as both middleware or
       direct via a method call.
    """
    
    def __init__(self, config, app=None):
        """
        Takes a config-instance and optionally a WSGI-app if this will be used as a 
        middleware. Raises `InitFailure` if the spool_path does not exist in the 
        config or cannot be created.
        
        - `config`: a `benri.config.Config` instance.
        - `app`: a WSGI application, if None this cannot be used as a WSGI-
                 middleware.
                 
        Usage:
        
        >>> spool = Spool({'service': {'spool_path': '/tmp/'}})
        >>> f_name = '/etc/hosts'
        >>> f = open(f_name, 'r')
        >>> tmp = spool.dump(f, os.path.getsize(f_name))
        >>> #''.join([l for l in open(tmp)])
        >>> os.unlink(tmp)
        """
        try:
            spool_path = config['service']['spool_path']
        except KeyError, e:
            raise InitFailure("spool_path could not be found in the [service] section in the config file")
                        
        if not os.path.exists(spool_path):
            try:
                os.mkdir(spool_path)
            except Exception, e:
                raise InitFailure("Could not create spool_path: %s" %(spool_path))

        self.__spool_path = spool_path
        self.__app = app        

    def __call__(self, env, resp):
        """
        When used as a WSGI-middleware, the input-data is read into a spool-file.
        The absolute-path to the file is passed to the application in the 
        environment-variable `benri.spool_file`.
        """
        
        in_file = env.get('wsgi.input', None)
        size = int(env.get('CONTENT_LENGTH', '-1'))
        
        env['benri.spool_file'] = self.dump(in_file, size)
        
        return self.__app(env, resp)
        

    def dump(self, file_obj, size=-1):
        """
        Dumps the contents of `file_obj` to a file in the spool-directory. Returns
        the absolute path to the new file.
        
        - `file_obj`: a file-like object with a `read`-method
        - `size`: number of bytes to be read from the file. If this is -1, we read
                  data until an EOF is found or StopIteration is raised.
        """        
        (fd, tmp_name) = mkstemp(dir=self.__spool_path)

        f = os.fdopen(fd, 'w+')
        for chunk in _FileIter(file_obj, size=size, block_size=2**14):
            f.write(chunk)
        
        f.close()
        
        return tmp_name
                
if __name__ == '__main__':
    import doctest
    doctest.testmod()    
